"""Implementation of XBAreaction class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import re
import sbmlxdf
from xbanalysis.utils.utils import get_species_stoic, get_local_params, expand_kineticlaw
from xbanalysis.utils.utils import xml_rba_ns
from xbanalysis.model.sbo_term import SboTerm


class XbaReaction:

    def __init__(self, s_reaction, species, functions, compartments):
        self.id = s_reaction.name
        self.name = s_reaction.get('name', self.id)
        if 'sboterm' in s_reaction:
            self.sboterm = SboTerm(s_reaction['sboterm'])
        self.reaction_string = s_reaction['reactionString']
        self.reversible = s_reaction['reversible']
        self.reactants = get_species_stoic(s_reaction['reactants'])
        self.products = get_species_stoic(s_reaction['products'])
        self.metabs_only = True
        for sid in list(self.reactants) + list(self.products):
            if species[sid].sboterm.is_in_branch('SBO:0000247') is False:
                self.metabs_only = False
        if 'modifiers' in s_reaction:
            self.modifiers = get_species_stoic(s_reaction['modifiers'])
        else:
            self.modifiers = {}
        if 'kineticLaw' in s_reaction:
            self.local_params = get_local_params(s_reaction)
            self.kinetic_law = s_reaction['kineticLaw'] if type(s_reaction['kineticLaw']) == str else ''
            # convert to numpy, replace local parameters with numerical values, inline lambda functions:
            expanded_kl = sbmlxdf.misc.mathml2numpy(self.kinetic_law)
            for pid, val in self.local_params.items():
                expanded_kl = re.sub(r'\b' + pid + r'\b', str(val), expanded_kl)
            self.expanded_kl = expand_kineticlaw(expanded_kl, functions)
            for cid in compartments.keys():
                if re.search(r'\b'+cid+r'\b', self.expanded_kl) is not None:
                    self.compartment = cid
        # retrieve optional flux balance bounds, ensuring consistent bounds
        if 'fbcLb' in s_reaction and 'fbcUb' in s_reaction:
            self.fbc_lb = s_reaction['fbcLb']
            self.fbc_ub = max(self.fbc_lb, s_reaction['fbcUb'])
            if self.reversible is False:
                self.fbc_lb = max(0, self.fbc_lb)
                self.fbc_ub = max(0, self.fbc_ub)
        if 'fbcGeneProdAssoc' in s_reaction:
            if type(s_reaction['fbcGeneProdAssoc']) == str:
                self.fbc_gpa = re.sub(r'^assoc=', '', s_reaction['fbcGeneProdAssoc'])
        if 'xmlAnnotation' in s_reaction:
            attrs = sbmlxdf.extract_xml_attrs(s_reaction['xmlAnnotation'], ns=xml_rba_ns, token='kinetics')
            if 'kapp_f_per_s' in attrs:
                self.rba_kapp_f = float(attrs['kapp_f_per_s'])
            if 'kapp_r_per_s' in attrs:
                self.rba_kapp_r = float(attrs['kapp_r_per_s'])
            if 'Km_ext_M' in attrs:
                self.rba_km_ext = float(attrs['Km_ext_M'])
        self.ps_product = ''
        self.enzyme = None

    def set_enzyme(self, species):
        for sid in self.modifiers:
            if species[sid].sboterm.is_in_branch('SBO:0000246'):
                self.enzyme = sid
                break

    def add_species_usage(self, species):
        # for metabolic reactions add to the species consumed/produced the metabolic reaction id
        # !!! we could also enter the Reaction Object references instead
        if self.sboterm.is_in_branch('SBO:0000167'):
            for sid, val in self.products.items():
                species[sid].add_m_reaction(self.id, val)
            for sid, val in self.reactants.items():
                species[sid].add_m_reaction(self.id, -val)
        # protein degradation reactions
        if self.sboterm.is_in_branch('SBO:0000179'):
            for sid, val in self.reactants.items():
                if species[sid].sboterm.is_in_branch('SBO:0000246'):
                    species[sid].add_p_degrad_rid(self.id, -val)
        # protein transitions
        if self.sboterm.is_in_branch('SBO:0000182'):
            for sid, val in self.reactants.items():
                if species[sid].sboterm.is_in_branch('SBO:0000246'):
                    species[sid].add_p_transition_rid(self.id, -val)
            for sid, val in self.products.items():
                if species[sid].sboterm.is_in_branch('SBO:0000246'):
                    species[sid].add_p_transition_rid(self.id, val)

        # protein synthesis reactions
        if self.sboterm.is_in_branch('SBO:0000205'):
            for sid, val in self.products.items():
                if species[sid].sboterm.is_in_branch('SBO:0000246'):
                    self.ps_product = sid
                    species[sid].set_ps_rid(self.id)
                else:
                    species[sid].add_ps_reaction(self.id, val)
            for sid, val in self.reactants.items():
                species[sid].add_ps_reaction(self.id, -val)
